﻿using System.Collections.Generic;
using System.Linq;

namespace Sharing
{
	public class Node
	{
		public enum NodeType { App, Lam, Var };
		public readonly NodeType Type;

		public Node Canonic = null;
		public bool Building = false;

		public Node Left { get; private set; } = null;
		public Node Right { get; private set; } = null;
		public Node Body { get { return Left; } set { Left = value; } }
		public Node Binder { get { return Left; } set { Left = value; } }

		public List<Node> Parents = new List<Node>();

		public readonly List<UndirectedEdge> UndirectedEdges = new List<UndirectedEdge>();

		public Queue<Node> Queue = new Queue<Node>();

		public int Index;
		public string Label => Type == NodeType.App ? "@" : (Type == NodeType.Lam ? "λ" : "v");

		public Node(NodeType type) { Type = type; }

		public void AddLeft(Node node)
		{
			Left = node;
			node.Parents.Add(this);
		}

		public void AddRight(Node node)
		{
			Right = node;
			node.Parents.Add(this);
		}
		public void AddBody(Node node)
		{
			Body = node;
			node.Parents.Add(this);
		}

		public void AddUndirectedEdge(UndirectedEdge edge) => UndirectedEdges.Add(edge);

		public void Copy(Dictionary<Node, Node> map)
		{
			Node target = map[this];
			target.Left = Left == null ? null : map[Left];
			target.Right = Right == null ? null : map[Right];

			foreach (Node parent in Parents)
				target.Parents.Add(map[parent]);
		}

		public void RemoveLeft()
		{
			Left.Parents.Remove(this);
			Left = null;
		}

		public void RemoveRight()
		{
			Right.Parents.Remove(this);
			Right = null;
		}

		public bool HasAncestor(Node n)
		{
			if (this == n) return true;
			else return this.Parents.Exists(m => m.HasAncestor(n));
		}

		public bool DominatedBy(Node n)
		{
			if (this == n) return true;
                        else if (Parents.Count() == 0) return false;
			else return Parents.TrueForAll(m => m.DominatedBy(n));
		}

		public List<Node> DescendantVars()
		{
			switch (Type)
			{
				case NodeType.Var: return new List<Node>() { this };
				case NodeType.Lam: return new List<Node>();
				default: return Left.DescendantVars().Concat(Right.DescendantVars()).ToList();
			}
		}

	}

	public class UndirectedEdge
	{
		public Node Left { get; private set; }
		public Node Right { get; private set; }

		public UndirectedEdge(Node left, Node right)
		{
			Left = left;
			Right = right;
		}

		public Node FindTarget(Node source) 
		{
			if (source == Left)
				return Right;
			else if (source == Right)
				return Left;

			throw new ExecutionException("can't find target.");
		}
	}
}
